Attribute VB_Name = "HookHandler_Global"
Option Explicit

'==================================================================
'
' Hook Handler Control About Dialog
'
' 1 October 2001
'
' You're free in using this software, as long as the
' orignal copyright/author information remains intact.
' In other words, respecting the person(s) who made it
' possible for you to view/learn something of this software.
'
' Copyright  2001 - Diederik van der Boor
' http://www.codingdomain.com
'
'==================================================================


' This module contains some global
' API declarations (for calling winprocs)
' and a winproc for the redirected events
' (AddressOf operator can only get)
' (proc addresses from a basic-module)

Private Const MODULE_NAME As String = "Hooking.HookHandler_Global."
Private Const ERROR_INVALID_HOOK_FILTER = 1426 ' Invalid hook procedure type.
Private Const ERROR_INVALID_FILTER_PROC = 1427 ' Invalid hook procedure.
Private Const ERROR_HOOK_NEEDS_HMOD = 1428     ' Cannot set nonlocal hook without a module handle.
Private Const ERROR_GLOBAL_ONLY_HOOK = 1429    ' This hook procedure can only be set globally.
Private Const ERROR_JOURNAL_HOOK_SET = 1430    ' The journal hook procedure is already installed.
Private Const ERROR_HOOK_NOT_INSTALLED = 1431  ' The hook procedure is not installed.
Private Const ERROR_MOD_NOT_FOUND = 126        ' The specified module could not be found.

Private Type HookInfo
  HookObjPtr As Long
  HookType As HookingTypes
End Type

Public Const HOOK_FIRST_MSG = -1
Public Const HOOK_LAST_MSG = 12

Private CurhHook(HOOK_FIRST_MSG To HOOK_LAST_MSG) As Long      ' Attached hHook to type.
Public CurHooksCount(HOOK_FIRST_MSG To HOOK_LAST_MSG) As Long  ' Number of HookHandler Clients.
Private CurHookData() As HookInfo                              ' Information about hooks and their clients.
Private CurHookDataCount As Long                               ' Number of information entries.

Public Type POINTAPI
  X As Long
  Y As Long
End Type

Public Type MSG
  hWnd As Long
  message As Long
  wParam As Long
  lParam As Long
  time As Long
  pt As POINTAPI
End Type

' Set/Get the window proc handle
Private Declare Function SetWindowsHookEx Lib "user32" Alias "SetWindowsHookExA" (ByVal idHook As Long, ByVal lpFn As Long, ByVal hmod As Long, ByVal dwThreadId As Long) As Long
Private Declare Function UnhookWindowsHookEx Lib "user32" (ByVal hHook As Long) As Long
Private Declare Function CallNextHookEx Lib "user32" (ByVal hHook As Long, ByVal nCode As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
Private Declare Function GetCurrentThreadId Lib "kernel32" () As Long

' Pointers
Public Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (pDest As Any, pSrc As Any, ByVal ByteLen As Long)
Public Declare Function IsBadReadPtr Lib "kernel32" (lp As Any, ByVal ucb As Long) As Long
Public Declare Function IsBadWritePtr Lib "kernel32" (lp As Any, ByVal ucb As Long) As Long
Public Declare Function IsBadCodePtr Lib "kernel32" (lpFn As Any) As Long
'Public Declare Function IsBadStringPtr Lib "kernel32" (lpsz As Any, ByVal ucchMax As Long) As Long

' Returns the address of a procedure passed by the AddressOf operator
Public Function ProcToPtr(ByVal AddressOf_Proc As Long) As Long
  ProcToPtr = AddressOf_Proc ' Why do we have to write nonesence like this? (Can't be just have an "Function(Pointer)" datatype?)
End Function

' Use object pointers to prevent VB from incrementing the class instance reference count
Public Function ObjFromPtr(ByVal ObjectPtr As Long) As Object
  Dim Obj As Object
  Dim ObjGet As Object                  ' Receiver for pointer
  
  ' These safety checks aren't 100% ok,
  ' but it will catch some really invalid pointers!
  If Not (CBool(IsBadCodePtr(ByVal ObjectPtr)) And CBool(IsBadReadPtr(ByVal ObjectPtr, 4)) And CBool(IsBadWritePtr(ByVal ObjectPtr, 4))) Then
    On Error Resume Next
      Call CopyMemory(ObjGet, ObjectPtr, 4) ' Turn the pointer into an ILLEGAL, UNCOUNTED interface
      Set Obj = ObjGet                      ' Assign to legal reference (Object references will be counted and registered by VB)
      Call CopyMemory(ObjGet, 0&, 4)        ' Destroy the illegal reference (Can't harm VB anymore)
    On Error GoTo 0
    Set ObjFromPtr = Obj                  ' Return legal object reference
  End If
End Function

' First, we insert a new hook, which executed before the message loop of this thread!
' We must return control of the messages back to windows before the program exits
Public Sub AttachHook(ByVal HookObjPtr As Long, ByVal HookType As HookingTypes)
  Debug.Assert Not (CurHooksCount(HookType) = 0 Xor CurhHook(HookType) = 0)
  ' If have a hook function: Add the class to the hook receive list,
  ' Else we create the hook here.
  If CurhHook(HookType) = 0 Then
    Dim HookCallBackAddr As Long
    Select Case HookType
      Case whkMessageFilter:      HookCallBackAddr = ProcToPtr(AddressOf HookHandler_MessageProc)
'     Case whkSysMessageFilter:   HookCallBackAddr = ProcToPtr(AddressOf HookHandler_SysMsgProc)
      Case whkMsgPoolReading:     HookCallBackAddr = ProcToPtr(AddressOf HookHandler_GetMsgProc)
      Case whkWinProcCalls:       HookCallBackAddr = ProcToPtr(AddressOf HookHandler_CallWndProc)
      Case whkWinProcCallsRetVal: HookCallBackAddr = ProcToPtr(AddressOf HookHandler_CallWndRetProc)
      Case whkMouseMessages:      HookCallBackAddr = ProcToPtr(AddressOf HookHandler_MouseProc)
      Case whkKeyboardStrokes:    HookCallBackAddr = ProcToPtr(AddressOf HookHandler_KeyboardProc)
      Case whkUITraining:         HookCallBackAddr = ProcToPtr(AddressOf HookHandler_CBTProc)
      Case whkShellNotifications: HookCallBackAddr = ProcToPtr(AddressOf HookHandler_ShellProc)
      Case whkForegroundAppIdle:  HookCallBackAddr = ProcToPtr(AddressOf HookHandler_ForegroundIdleProc)
      Case whkJournalPlayBack:    HookCallBackAddr = ProcToPtr(AddressOf HookHandler_JournalPlaybackProc)
      Case whkJournalRecord:      HookCallBackAddr = ProcToPtr(AddressOf HookHandler_JournalRecordProc)
      Case Else:                  DIE whkErrUnsupportedType, MODULE_NAME & "AttachHook(" & HookObjPtr & ", " & HookType & ")"
    End Select
    ' External callbacks aren't supported, so hMod = 0, dwThreadID = GetCurrentThreadId() and HookCallBackAddr is a local address.
    CurhHook(HookType) = SetWindowsHookEx(HookType, HookCallBackAddr, 0, GetCurrentThreadId())
    If CurhHook(HookType) = 0 Then
      Dim ErrStr As String
      Select Case Err.LastDllError
        Case 0:                         ErrStr = vbNullString
        Case ERROR_INVALID_HOOK_FILTER: ErrStr = "Invalid hook procedure type."
        Case ERROR_INVALID_FILTER_PROC: ErrStr = "Invalid hook procedure."
        Case ERROR_HOOK_NEEDS_HMOD:     ErrStr = "Cannot set nonlocal hook without a module handle."
        Case ERROR_GLOBAL_ONLY_HOOK:    ErrStr = "This hook procedure can only be set globally."
        Case ERROR_JOURNAL_HOOK_SET:    ErrStr = "The journal hook procedure is already installed."
        Case ERROR_HOOK_NOT_INSTALLED:  ErrStr = "The hook procedure is not installed."
        Case ERROR_MOD_NOT_FOUND:       ErrStr = "The specified module could not be found."
        Case Else:                      ErrStr = "Out of memory or security issue prevents hooking"
      End Select
      DIE whkErrHookingFailed, MODULE_NAME & "AttachHook(" & HookObjPtr & ", " & HookType & ")", ErrStr
    End If
  Else
    If HookType = whkUITraining Or HookType = whkJournalPlayBack Or HookType = whkJournalRecord Then
      DIE whkErrCantShareHook, MODULE_NAME & "AttachHook(" & HookObjPtr & ", " & HookType & ")"
    End If
  End If
  CurHooksCount(HookType) = CurHooksCount(HookType) + 1
  If CurHookDataCount = 0 Then
    CurHookDataCount = CurHookDataCount + 1
    ReDim CurHookData(1 To CurHookDataCount) As HookInfo
  Else
    If GetHook(HookObjPtr, HookType) > 0 Then DIE whkErrHooked, MODULE_NAME & "AttachHook(" & HookObjPtr & ", " & HookType & ")"
    CurHookDataCount = CurHookDataCount + 1
    ReDim Preserve CurHookData(1 To CurHookDataCount) As HookInfo
  End If
  CurHookData(CurHookDataCount).HookObjPtr = HookObjPtr
  CurHookData(CurHookDataCount).HookType = HookType
End Sub

Public Sub DetachHook(ByVal HookObjPtr As Long, ByVal HookType As HookingTypes)
  Dim I As Long
  Dim HookIndex As Long
  
  HookIndex = GetHook(HookObjPtr, HookType)
  
  If HookIndex Then
    ' Unhook the hook
    CurHooksCount(HookType) = CurHooksCount(HookType) - 1
    If CurHooksCount(HookType) = 0 Then
      If UnhookWindowsHookEx(CurhHook(HookType)) = 0 Then DIE whkErrUnHookingFailed, MODULE_NAME & "DetachHook(" & HookObjPtr & ", " & HookType & ")"
      CurhHook(HookType) = 0
    ElseIf CurHooksCount(HookType) < 0 Then
      Debug.Assert UnhookWindowsHookEx(CurhHook(HookType))
      CurhHook(HookType) = 0
      DIE whkErrInternal, MODULE_NAME & "DetachHook(" & HookObjPtr & ", " & HookType & ")"
    End If
    
    ' Remove other data from arrays
    For I = HookIndex + 1 To CurHookDataCount
      LSet CurHookData(I - 1) = CurHookData(I)
    Next
    CurHookDataCount = CurHookDataCount - 1
    If CurHookDataCount > 0 Then
      ReDim Preserve CurHookData(1 To CurHookDataCount) As HookInfo
'    Else
'      Erase CurHookData
    End If
    Exit Sub
  End If
  
  DIE whkErrUnHookingFailed, MODULE_NAME & "DetachHook(" & HookObjPtr & ", " & HookType & ")"
End Sub

Public Function GetHook(ByVal HookObjPtr As Long, ByVal HookType As HookingTypes) As Long
  Dim I As Long
  For I = 1 To CurHookDataCount
    If CurHookData(I).HookType = HookType Then
      If CurHookData(I).HookObjPtr = HookObjPtr Then
        GetHook = I
      End If
    End If
  Next
End Function





Public Function HookHandler_MessageProc(ByVal nCode As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
  Dim hHook As Long
  hHook = CurhHook(whkMessageFilter)
  If nCode >= 0 Then
    Dim I As Long
    Dim HookObj As HookHandler
    For I = CurHookDataCount To 1 Step -1
      If CurHookData(I).HookType = whkMessageFilter Then
        On Error Resume Next
          Set HookObj = ObjFromPtr(CurHookData(I).HookObjPtr)
        On Error GoTo 0
        If Not HookObj Is Nothing Then
          If HookObj.RH_MessageProc(nCode, wParam, lParam) Then
            HookHandler_MessageProc = 1
            Exit Function
          End If
        End If
      End If
    Next
  End If
  
  ' Return this when nCode < 0 or when message not consumed
  HookHandler_MessageProc = CallNextHookEx(hHook, nCode, wParam, lParam)
End Function

Public Function HookHandler_GetMsgProc(ByVal nCode As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
  Dim hHook As Long
  hHook = CurhHook(whkMsgPoolReading)
  If nCode >= 0 Then
    Dim I As Long
    Dim HookObj As HookHandler
    For I = CurHookDataCount To 1 Step -1
      If CurHookData(I).HookType = whkMsgPoolReading Then
        On Error Resume Next
          Set HookObj = ObjFromPtr(CurHookData(I).HookObjPtr)
        On Error GoTo 0
        If Not HookObj Is Nothing Then
          If HookObj.RH_GetMsgProc(nCode, wParam, lParam) Then Exit Function
        End If
      End If
    Next
  End If
  
  ' Return this when nCode < 0 or when message not consumed
  HookHandler_GetMsgProc = CallNextHookEx(hHook, nCode, wParam, lParam)
End Function

Public Function HookHandler_CallWndProc(ByVal nCode As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
  Dim hHook As Long
  hHook = CurhHook(whkWinProcCalls)
  If nCode >= 0 Then
    Dim I As Long
    Dim HookObj As HookHandler
    For I = CurHookDataCount To 1 Step -1
      If CurHookData(I).HookType = whkWinProcCalls Then
        On Error Resume Next
          Set HookObj = ObjFromPtr(CurHookData(I).HookObjPtr)
        On Error GoTo 0
        If Not HookObj Is Nothing Then
          If HookObj.RH_CallWndProc(nCode, wParam, lParam) Then Exit Function
        End If
      End If
    Next
  End If
  
  ' Return this when nCode < 0 or when message not consumed
  HookHandler_CallWndProc = CallNextHookEx(hHook, nCode, wParam, lParam)
End Function

Public Function HookHandler_CallWndRetProc(ByVal nCode As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
  Dim hHook As Long
  hHook = CurhHook(whkWinProcCallsRetVal)
  If nCode >= 0 Then
    Dim I As Long
    Dim HookObj As HookHandler
    For I = CurHookDataCount To 1 Step -1
      If CurHookData(I).HookType = whkWinProcCallsRetVal Then
        On Error Resume Next
          Set HookObj = ObjFromPtr(CurHookData(I).HookObjPtr)
        On Error GoTo 0
        If Not HookObj Is Nothing Then
          If HookObj.RH_CallWndRetProc(nCode, wParam, lParam) Then Exit Function
        End If
      End If
    Next
  End If
  ' Return this when nCode < 0 or when message not consumed
  HookHandler_CallWndRetProc = CallNextHookEx(hHook, nCode, wParam, lParam)
End Function

Public Function HookHandler_MouseProc(ByVal nCode As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
  Dim hHook As Long
  hHook = CurhHook(whkMouseMessages)
  If nCode >= 0 Then
    Dim I As Long
    Dim HookObj As HookHandler
    For I = CurHookDataCount To 1 Step -1
      If CurHookData(I).HookType = whkMouseMessages Then
        On Error Resume Next
          Set HookObj = ObjFromPtr(CurHookData(I).HookObjPtr)
        On Error GoTo 0
        If Not HookObj Is Nothing Then
          If HookObj.RH_MouseProc(nCode, wParam, lParam) Then
            HookHandler_MouseProc = 1
            Exit Function
          End If
        End If
      End If
    Next
  End If
  
  ' Return this when nCode < 0 or when message not consumed
  HookHandler_MouseProc = CallNextHookEx(hHook, nCode, wParam, lParam)
End Function

Public Function HookHandler_KeyboardProc(ByVal nCode As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
  Dim hHook As Long
  hHook = CurhHook(whkKeyboardStrokes)
  If nCode >= 0 Then
    Dim I As Long
    Dim HookObj As HookHandler
    For I = CurHookDataCount To 1 Step -1
      If CurHookData(I).HookType = whkKeyboardStrokes Then
        On Error Resume Next
          Set HookObj = ObjFromPtr(CurHookData(I).HookObjPtr)
        On Error GoTo 0
        If Not HookObj Is Nothing Then
          If HookObj.RH_KeyboardProc(nCode, wParam, lParam) Then
            HookHandler_KeyboardProc = 1
            Exit Function
          End If
        End If
      End If
    Next
  End If
  
  ' Return this when nCode < 0 or when message not consumed
  HookHandler_KeyboardProc = CallNextHookEx(hHook, nCode, wParam, lParam)
End Function

Public Function HookHandler_CBTProc(ByVal nCode As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
' Return value is ignored for: HCBT_CLICKSKIPPED, HCBT_KEYSKIPPED, HCBT_QS
  Dim hHook As Long
  hHook = CurhHook(whkUITraining)
  If nCode >= 0 Then
    Dim I As Long
    Dim HookObj As HookHandler
    ' Find the hook control, and don't look further
    ' It's logical that only one Computer-Based-Training program is running in
    ' the current thread, and that it swallows all other messages!
    ' If you don't like it, make your own CBT_Proc and call it with SetWindowsHookEx!
    For I = CurHookDataCount To 1 Step -1
      If CurHookData(I).HookType = whkUITraining Then
        On Error Resume Next
          Set HookObj = ObjFromPtr(CurHookData(I).HookObjPtr)
        On Error GoTo 0
        If Not HookObj Is Nothing Then
          If HookObj.RH_CBTProc(nCode, wParam, lParam) Then
            HookHandler_CBTProc = 0 ' Allow the action
          Else
            HookHandler_CBTProc = 1 ' Prevent the action
          End If
          Exit For
        End If
      End If
    Next
  Else
    ' Return this when nCode < 0
    HookHandler_CBTProc = CallNextHookEx(hHook, nCode, wParam, lParam)
  End If
End Function

Public Function HookHandler_ShellProc(ByVal nCode As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
  Dim hHook As Long
  hHook = CurhHook(whkShellNotifications)
  If nCode >= 0 Then
    Dim I As Long
    Dim HookObj As HookHandler
    For I = CurHookDataCount To 1 Step -1
      If CurHookData(I).HookType = whkShellNotifications Then
        On Error Resume Next
          Set HookObj = ObjFromPtr(CurHookData(I).HookObjPtr)
        On Error GoTo 0
        If Not HookObj Is Nothing Then
          Call HookObj.RH_ShellProc(nCode, wParam, lParam) ' Notify only.
        End If
        ' return 0, or 1 if nCode = HSHELL_APPCOMMAND and WM_COMMAND is processed
      End If
    Next
  Else
    ' Return this when nCode < 0
    HookHandler_ShellProc = CallNextHookEx(hHook, nCode, wParam, lParam)
  End If
End Function

Public Function HookHandler_ForegroundIdleProc(ByVal nCode As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
  Dim hHook As Long
  hHook = CurhHook(whkForegroundAppIdle)
  If nCode >= 0 Then
    Dim I As Long
    Dim HookObj As HookHandler
    For I = CurHookDataCount To 1 Step -1
      If CurHookData(I).HookType = whkForegroundAppIdle Then
        On Error Resume Next
          Set HookObj = ObjFromPtr(CurHookData(I).HookObjPtr)
        On Error GoTo 0
        If Not HookObj Is Nothing Then
          If HookObj.RH_ForegroundIdleProc(nCode, wParam, lParam) Then Exit Function
        End If
      End If
    Next
  End If
  ' Return this when nCode < 0 or when message not consumed
  HookHandler_ForegroundIdleProc = CallNextHookEx(hHook, nCode, wParam, lParam)
End Function

Public Function HookHandler_JournalPlaybackProc(ByVal nCode As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
  Dim hHook As Long
  hHook = CurhHook(whkJournalPlayBack)
  If nCode >= 0 Then
    Dim I As Long
    Dim HookObj As HookHandler
    For I = CurHookDataCount To 1 Step -1
      If CurHookData(I).HookType = whkJournalPlayBack Then
        On Error Resume Next
          Set HookObj = ObjFromPtr(CurHookData(I).HookObjPtr)
        On Error GoTo 0
        If Not HookObj Is Nothing Then
          HookHandler_JournalPlaybackProc = HookObj.RH_JournalPlaybackProc(nCode, wParam, lParam)
          Exit Function ' Return waiting time if HC_GETNEXT?
        End If
      End If
    Next
  Else
    ' Return this when nCode < 0
    HookHandler_JournalPlaybackProc = CallNextHookEx(hHook, nCode, wParam, lParam)
  End If
End Function

Public Function HookHandler_JournalRecordProc(ByVal nCode As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
  Dim hHook As Long
  hHook = CurhHook(whkJournalRecord)
  If nCode >= 0 Then
    Dim I As Long
    Dim HookObj As HookHandler
    For I = CurHookDataCount To 1 Step -1
      If CurHookData(I).HookType = whkJournalRecord Then
        On Error Resume Next
          Set HookObj = ObjFromPtr(CurHookData(I).HookObjPtr)
        On Error GoTo 0
        If Not HookObj Is Nothing Then
          Call HookObj.RH_JournalRecordProc(nCode, wParam, lParam)
          Exit Function ' RetVal is ignored.
        End If
      End If
    Next
  Else
    ' Return this when nCode < 0
    HookHandler_JournalRecordProc = CallNextHookEx(hHook, nCode, wParam, lParam)
  End If
End Function

'Public Function HookHandler_SysMsgProc(ByVal nCode As Long, ByVal wParam As Long, ByVal lParam As Long) As Long
' This callback needs to be an API in a systemwide used DLL.
'End Function


Public Sub DIE(ErrorCode As HookingErrors, Optional Source As String = MODULE_NAME & "DIE(...)", Optional Comments As String)
  Dim ErrorStr As String
  Select Case ErrorCode
    Case whkErrInternal:
    Case whkErrHooked:          ErrorStr = "The hook is already available form this control!"
    Case whkErrNotHooked:       ErrorStr = "The hook has already been destroyed!"
    Case whkErrHookingFailed:   ErrorStr = "The attempt to hook certain system events failed!"
    Case whkErrUnHookingFailed: ErrorStr = "The attempt to unhook a certain hook failed!"
    Case whkErrUnsupportedType: ErrorStr = "The specified hook isn't supported, properly caused by the fact that:" & vbCrLf _
                                         & vbTab & "- You entered a bad constant." & vbCrLf _
                                         & vbTab & "- The hook type didn't exist when this version case out." & vbCrLf _
                                         & vbTab & "- Systemwide hooks and hooks for threads in different applications must reside in DLLs / API's!"
    Case whkErrCantShareHook:   ErrorStr = "This hook type can't be shared by multiple control instances in this thread!"
    Case Else:                  ErrorStr = Error(ErrorCode)
  End Select
  If ErrorStr = "" Or ErrorStr = "Application-defined or object-defined error" Then ErrorStr = "Unexpected Error! (code " & ErrorCode & ")"
  Err.Raise ErrorCode, Source, ErrorStr & IIf(Len(Comments), vbCrLf & Comments, vbNullString)
End Sub
